from bofire.surrogates.botorch_surrogates import BotorchSurrogates
from bofire.surrogates.empirical import EmpiricalSurrogate
from bofire.surrogates.mapper import map
from bofire.surrogates.mixed_single_task_gp import MixedSingleTaskGPSurrogate
from bofire.surrogates.mlp import MLPEnsemble
from bofire.surrogates.random_forest import RandomForestSurrogate
from bofire.surrogates.single_task_gp import SingleTaskGPSurrogate
from bofire.surrogates.surrogate import Surrogate
from bofire.surrogates.trainable import TrainableSurrogate
from bofire.surrogates.values import PredictedValue
from bofire.surrogates.xgb import XGBoostSurrogate
