from bofire.data_models.strategies.strategy import Strategy


class SamplerStrategy(Strategy):
    """Base class for sampling methods in BoFire for sampling from constrained input spaces."""

    pass
